//
// QFleximon - Qt program to monitor fleximon daemons.
//
// This program is part of Fleximon, a flexible, plugin-based client-server
// monitoring framework.
//
// See http://fleximon.sourceforge.net for details.
//
// 2002, Patrick Schemitz <schemitz@users.sourceforge.net>
//

// C header
#include <sys/socket.h>
#include <arpa/inet.h>
#include <netdb.h>

// C++ header (be anal about namespaces)
#include <iostream>
#include <fstream>
using std::endl;
using std::vector;

// Qt header
#include <qdatastream.h>
#include <qvalidator.h>
#include <qtimer.h>
#include <qlabel.h>
#include <qapplication.h>
#include <qimage.h>
#include <qpixmap.h>
#include <qtoolbar.h>
#include <qtoolbutton.h>
#include <qpopupmenu.h>
#include <qmenubar.h>
#include <qtextedit.h>
#include <qlineedit.h>
#include <qcheckbox.h>
#include <qlayout.h>
#include <qfile.h>
#include <qfiledialog.h>
#include <qstatusbar.h>
#include <qmessagebox.h>
#include <qprinter.h>
#include <qapplication.h>
#include <qaccel.h>
#include <qtextstream.h>
#include <qpainter.h>
#include <qpaintdevicemetrics.h>
#include <qwhatsthis.h>
#include <qtooltip.h>
#include <qtabdialog.h>
#include <qinputdialog.h>

// fleximon header
#include "libfleximon.h"

#include "qfleximon.hh"


static const unsigned FMDefaultPort = 9000;
static const unsigned MaxQueryBufSize = 10240;


static unsigned max (unsigned a, unsigned b)
{
    return (a > b ? a : b);
}


//
// Toggle debug output. If you want it, make Debug an alias for
// (aka. a reference to) cout. To get rid of all the debug output,
// make Debug an ofstream writing to /dev/null.
//
std::ostream& Debug = std::cout;
//std::ofstream Debug ("/dev/null");


//
// Representation of one host
//
QFlexiHost::QFlexiHost (QIconView* parent, QString h, unsigned p)
    : QIconViewItem (parent), pic(32,32), hostname(h), port(p), delay(0)
{
    pic.fill(Qt::green);
    setPixmap(pic);
    setText(hostname);
    lookupHost();
    updateHost(true);
}

QFlexiHost* QFlexiHost::clone () const
{
    QFlexiHost* newHost = new QFlexiHost(iconView(),hostname,port);
    newHost->ip = ip;
    newHost->plugin = plugin;
    newHost->delay = delay;
    newHost->updateHost();
    return newHost;
}

QFlexiHost::~QFlexiHost ()
{
}

bool QFlexiHost::fmGetList ()
{
    char buffer [MaxQueryBufSize];
    Plugin pi;
    plugin.clear();
    if (fm_query(ip,port,"LIST",buffer,MaxQueryBufSize)) {
	Debug << "fleximond not responding on " << hostname << ":" << port
	      << endl;
	delay = DelayOnError;
	return false;
    }
    char* cmd = buffer;
    char* p = buffer;
    while (*p) {
	if ((*p) == ',') {
	    *p = 0;
	    pi.name = cmd;
	    pi.desc = "";
	    pi.value = "";
	    pi.args = "";
	    pi.use = false;
	    plugin.push_back(pi);
	    cmd = ++p;
	} else
	    p++;
    }
    pi.name = cmd;
    pi.desc = "";
    pi.value = "";
    pi.args = "";
    pi.use = false;
    plugin.push_back(pi);
    for (unsigned i=0; i<plugin.size(); i++) {
	if (fm_query(ip,port,QString("DESCRIBE ") + plugin[i].name,
		     buffer,MaxQueryBufSize)) {
	    Debug << "fleximond went down: " << hostname << ":" << port
		  << endl;
	    delay = DelayOnError;
	    return false;
	}
	plugin[i].desc = buffer;
	Debug << plugin[i].name << ": " << plugin[i].desc << endl;
    }
    return true;
}

bool QFlexiHost::fmGetValues ()
{
    char buffer [MaxQueryBufSize];
    for (unsigned i=0; i<plugin.size(); i++) {
	if (!plugin[i].use) continue;
	if (fm_query(ip,port,(plugin[i].name+" "+plugin[i].args),
		     buffer,MaxQueryBufSize)) {
	    Debug << "fleximond not responding on " << hostname << ":" << ip
		  << endl;
	    delay = DelayOnError;
	    return false;
	}
	plugin[i].value = buffer;
	Debug << plugin[i].name << " value: " << plugin[i].value << endl;
    }
    return true;
}

void QFlexiHost::updateHost (bool force)
{
    if (!force && delay > 0) {
	Debug << hostname << ":" << port << " was down, " << delay
	      << " updates to be skipped." << endl;
	delay--;
	return;
    }
    Debug << "Updating " << hostname << ":" << port << endl;
    if (plugin.size() == 0)
	if (!fmGetList()) {
	    pic.fill(Qt::red);
	    setPixmap(pic);
	    setSelected(false);
	    return;
	}
    if (!fmGetValues()) {
	pic.fill(Qt::red);
	setPixmap(pic);
	setSelected(false);
	return;
    }
    unsigned maxError = 0;
    for (unsigned i=0; i<plugin.size(); i++) {
	if (plugin[i].value.find("WARNING") == 0)
	    maxError = max(maxError,1);
	if (plugin[i].value.find("ALERT") == 0)
	    maxError = max(maxError,2);
    }
    switch (maxError) {
    case 0:
	pic.fill(Qt::green);
	break;
    case 1:
	pic.fill(Qt::yellow);
	break;
    case 2:
	pic.fill(Qt::red);
	break;
    default:
	pic.fill(Qt::gray);
    }
    setPixmap(pic);
    setSelected(false);
    // TODO: is it possible to make the values into a tooltip? As far
    // as I understand, this does not work for QIconViewItems...
}

void QFlexiHost::lookupHost ()
{
    struct hostent* hostinfo = gethostbyname(hostname);
    if (hostinfo)
	memcpy((void*)&ip,hostinfo->h_addr_list[0],sizeof(in_addr_t));
    else
	ip = 0xffffffff;
}

void QFlexiHost::configure ()
{
    QTabDialog* t = new QTabDialog();
    QFlexiHostStatus* s = new QFlexiHostStatus(this);
    QFlexiPluginConfig* c = new QFlexiPluginConfig(this);
    QFlexiHostEdit* e = new QFlexiHostEdit(this);
    t->setCaption(hostname);
    t->addTab(s,"Status");
    t->addTab(c,"Plugins");
    t->addTab(e,"Edit Host");
    t->setApplyButton("Apply");
    t->setCancelButton("Close");
    QObject::connect(t,SIGNAL(applyButtonPressed()),s,SLOT(apply()));
    QObject::connect(t,SIGNAL(applyButtonPressed()),c,SLOT(apply()));
    QObject::connect(t,SIGNAL(applyButtonPressed()),e,SLOT(apply()));
    t->show();
}


QDataStream& operator<< (QDataStream& out, const QFlexiHost& host)
{
    out << host.hostname << host.port << host.plugin.size();
    for (unsigned i=0; i<host.plugin.size(); i++) {
	out << host.plugin[i].name << host.plugin[i].desc
	    << host.plugin[i].args << (int)host.plugin[i].use;
    }
    return out;
}

QDataStream& operator>> (QDataStream& in, QFlexiHost& host)
{
    unsigned size;
    in >> host.hostname >> host.port >> size;
    host.lookupHost();
    for (unsigned i=0; i<size; i++) {
	QFlexiHost::Plugin p;
	in >> p.name >> p.desc >> p.args;
	int use;
	in >> use;
	p.use = (use != 0);
	host.plugin.push_back(p);
    }
    return in;
}



//
// Dialog for editing host settings (name, port)
//
QFlexiHostEdit::QFlexiHostEdit (QFlexiHost* h, QWidget* parent)
    : QWidget(parent), host(h)
{
    in_addr ia;
    ia.s_addr = host->ip;
    QValidator* portValidator = new QIntValidator(1,65535,this);
    hostName = new QLineEdit(this);
    hostName->setText(host->hostname);
    hostPort = new QLineEdit(this);
    hostPort->setText(QString::number(host->port));
    hostPort->setValidator(portValidator);
    hostIP = new QLabel(this);
    hostIP->setText(inet_ntoa(ia));
    QGridLayout* grid = new QGridLayout(this,3,2);
    grid->addWidget(new QLabel("Host name:",this),0,0);
    grid->addWidget(hostName,0,1);
    grid->addWidget(new QLabel("IP address:",this),1,0);
    grid->addWidget(hostIP,1,1);
    grid->addWidget(new QLabel("Port:",this),2,0);
    grid->addWidget(hostPort,2,1);
    grid->setColStretch(0,2);
    grid->setColStretch(1,3);
    apply();
}

QFlexiHostEdit::~QFlexiHostEdit ()
{
}

void QFlexiHostEdit::apply ()
{
    if (hostName->edited()) {
	host->setText(hostName->text());
	host->hostname = hostName->text();
	host->lookupHost();
	in_addr ia;
	ia.s_addr = host->ip;
	hostIP->setText(inet_ntoa(ia));
	hostName->setEdited(false);
    }
    if (hostPort->edited()) {
	bool ok = true;
	unsigned tmp = hostPort->text().toUInt(&ok);
	if (ok) {
	    host->port = tmp;
	} else {
	    hostPort->setText(QString::number(host->port));
	}
	hostPort->setEdited(false);
    }
}



//
// Widget to display current status of host (plugin output)
//
QFlexiHostStatus::QFlexiHostStatus (QFlexiHost* h, QWidget* parent)
    : QWidget(parent), host(h), n(0), grid(0)
{
    n = host->plugin.size();
    command = new QLabel* [n];
    value = new QLabel* [n];
    for (unsigned i=0; i<n; i++) {
	command[i] = 0;
	value[i] = 0;
    }
    grid = new QGridLayout(this,1,2);
    grid->addWidget(new QLabel("Command",this),0,0);
    grid->addWidget(new QLabel("Value",this),0,1);
    apply();
}

QFlexiHostStatus::~QFlexiHostStatus ()
{
    delete [] command;
    delete [] value;
}

void QFlexiHostStatus::apply ()
{
    for (unsigned i=0; i<n; i++) {
	delete command[i];
	delete value[i];
	command[i] = value[i] = 0;
    }
    unsigned line = 1;
    for (unsigned i = 0; i<host->plugin.size(); i++) {
	if (!host->plugin[i].use) continue;
	command[i] = new QLabel((host->plugin[i].name+" "+
				 host->plugin[i].args),this);
	value[i] = new QLabel(host->plugin[i].value,this);
	grid->addWidget(command[i],line,0);
	grid->addWidget(value[i],line,1);
	line++;
    }
    Debug << line << " entries." << endl;
    grid->setColStretch(0,5);
    grid->setColStretch(1,10);
}



//
// Dialog to configure plugins on host
//
QFlexiPluginConfig::QFlexiPluginConfig (QFlexiHost* h, QWidget* parent)
    : QWidget(parent), host(h), n(0)
{
    n = host->plugin.size();
    checkbox = new QCheckBox* [n];
    arg = new QLineEdit* [n];
    QGridLayout* grid = new QGridLayout(this,n+1,4);
    grid->addWidget(new QLabel("Active",this),0,0);
    grid->addWidget(new QLabel("Command",this),0,1);
    grid->addWidget(new QLabel("Description",this),0,2);
    grid->addWidget(new QLabel("Arguments",this),0,3);
    for (unsigned i = 0; i<n; i++) {
	checkbox[i] = new QCheckBox(this);
	checkbox[i]->setChecked(host->plugin[i].use);
	arg[i] = new QLineEdit(this);
	arg[i]->setText(host->plugin[i].args);
	grid->addWidget(checkbox[i],1+i,0);
	grid->addWidget(new QLabel(host->plugin[i].name,this),1+i,1);
	grid->addWidget(new QLabel(host->plugin[i].desc,this),1+i,2);
	grid->addWidget(arg[i],1+i,3);
    }
    grid->setColStretch(0,1);
    grid->setColStretch(1,10);
    grid->setColStretch(2,20);
    grid->setColStretch(3,20);
}

QFlexiPluginConfig::~QFlexiPluginConfig ()
{
    delete [] checkbox;
    delete [] arg;
}

void QFlexiPluginConfig::apply ()
{
    for (unsigned i=0; i<n; i++) {
	host->plugin[i].use = checkbox[i]->isChecked();
	host->plugin[i].args = arg[i]->text();
    }
    host->updateHost();
}



//
// Application window class
//
QFlexiWindow::QFlexiWindow (QWidget* parent)
    : QMainWindow(parent,"QFlexiMon Window",WDestructiveClose)
{
    iv = new QIconView (this);
    iv->setFocus();
    setCentralWidget(iv);
    connect(iv,SIGNAL(contextMenuRequested(QIconViewItem*,const QPoint&)),
	    this,SLOT(hostMenu(QIconViewItem*, const QPoint&)));

    QPopupMenu* fileMenu = new QPopupMenu(this);
    menuBar()->insertItem("&File",fileMenu);
    fileMenu->insertItem("Open",this,SLOT(fileOpen()));
    fileMenu->insertItem("Save",this,SLOT(fileSave()));
    fileMenu->insertItem("Save as...",this,SLOT(fileSaveAs()));
    fileMenu->insertItem("New",this,SLOT(fileNew()));
    fileMenu->insertSeparator();
    fileMenu->insertItem("&Quit",qApp,SLOT(closeAllWindows()),CTRL+Key_Q);

    QPopupMenu* hostMenu = new QPopupMenu(this);
    menuBar()->insertItem("&Host",hostMenu);
    hostMenu->insertItem("New host",this,SLOT(hostNew()));
    hostMenu->insertItem("Clone host",this,SLOT(hostClone()));
    hostMenu->insertItem("Edit host",this,SLOT(hostMenu()));
    hostMenu->insertItem("Update host",this,SLOT(hostUpdate()));
    hostMenu->insertItem("Delete host",this,SLOT(hostDelete()));
    hostMenu->insertSeparator();
    hostMenu->insertItem("Arrange hosts",iv,SLOT(arrangeItemsInGrid()));

    updateMenu = new QPopupMenu(this);
    menuBar()->insertItem("Update",updateMenu);
    idAutoUpdate = updateMenu->insertItem("&Auto-update",this,
					  SLOT(toggleAutoUpdate()));
    updateMenu->setItemChecked(idAutoUpdate,false);
    updateMenu->insertItem("Update Interval",this,
			   SLOT(configUpdateInterval()));
    updateMenu->insertSeparator();
    updateMenu->insertItem("&Update",this,SLOT(update()));
    menuBar()->insertSeparator();

    QPopupMenu* help = new QPopupMenu(this);
    menuBar()->insertItem("&Help",help);
    help->insertItem("&About",this,SLOT(about()),Key_F1);
    help->insertItem("About &Qt",this,SLOT(aboutQt()));

    hostlistChanged = true;

    updateInterval = 5;
    timer = new QTimer(this);
    connect(timer,SIGNAL(timeout()),SLOT(autoUpdate()));
    timer->start(UpdateTimerInterval);
  
    statusBar()->message("Ready",2000);
    resize(450,600);
}

QFlexiWindow::~QFlexiWindow ()
{
    delete iv;
}

void QFlexiWindow::fileNew ()
{
    iv->clear();
    filename = "unnamed.qfm";
}

void QFlexiWindow::fileOpen ()
{
    char* buffer;
    unsigned version, n;
    QString tmp;
    tmp = QFileDialog::getOpenFileName(filename);
    if (tmp.isNull())
	return;
    filename = tmp;
    QFile f (filename);
    f.open(IO_ReadOnly);
    QDataStream s (&f);
    s >> buffer >> version;
    if (QString(buffer) != "QFlexiMon") {
	Debug << "Not a QFlexiMon file: " << buffer << endl;
	QMessageBox::critical(this,"Open failed",
			      "Not a QFlexiMon file: " + filename,"Ok");
	return;
    }
    if (version != 100) {
	Debug << "Wrong version: " << version << endl;
	QMessageBox::critical(this,"Open failed",
			      "File " + filename + " has " +
			      "wrong version: " +
			      QString::number(version),"Ok");
	return;
    }
    iv->clear();
    s >> n;
    Debug << "Reading " << n << "hosts..." << endl;
    for (unsigned i=0; i<n; i++) {
	QFlexiHost* host = new QFlexiHost(iv,"127.0.0.1",FMDefaultPort);
	s >> *host;
	iv->insertItem(host);
    }
    Debug << "Done reading." << endl;
}

void QFlexiWindow::fileSave ()
{
    if (filename == "unnamed.qfm") {
	fileSaveAs();
	return;
    }
    QFile f (filename);
    f.open(IO_WriteOnly);
    QDataStream s (&f);
    s << "QFlexiMon" << (unsigned)100;
    s << (unsigned)iv->count();
    QIconViewItem* item = iv->firstItem();
    while (item) {
	QFlexiHost* host = (QFlexiHost*)item;
	s << *host;
	item = item->nextItem();
    }
}

void QFlexiWindow::fileSaveAs ()
{
    QString tmp;
    tmp = QFileDialog::getSaveFileName(filename);
    if (tmp.isNull() || tmp == "unnamed.qfm")
	return;
    filename = tmp;
    fileSave();
}

void QFlexiWindow::toggleAutoUpdate ()
{
    updateMenu->setItemChecked(idAutoUpdate,
			       !updateMenu->isItemChecked(idAutoUpdate));
}

void QFlexiWindow::configUpdateInterval ()
{
    bool ok = true;
    timer->stop();
    int interval = QInputDialog::getInteger("Update Interval",
					    "Update interval in seconds",
					    updateInterval,
					    1,600,1,&ok,this);
    if (ok) {
	updateInterval = interval;
    }
    timer->start(UpdateTimerInterval);
}

void QFlexiWindow::hostNew ()
{
    QFlexiHost* host = new QFlexiHost(iv,"127.0.0.1",FMDefaultPort);
    host->configure();
}

void QFlexiWindow::hostClone ()
{
    QFlexiHost* current = (QFlexiHost*)iv->currentItem();
    QFlexiHost* host = current->clone();
    host->configure();
}

void QFlexiWindow::hostDelete ()
{
    QFlexiHost* current = (QFlexiHost*)iv->currentItem();
    int action = QMessageBox::warning(this,"Delete Host?",
				      QString("Really remove host ") +
				      current->text() + QString("?"),
				      QMessageBox::Ok, QMessageBox::Cancel);
    if (action == QMessageBox::Ok) {
	delete current;
    }
}

void QFlexiWindow::hostUpdate ()
{
    QFlexiHost* current = (QFlexiHost*)iv->currentItem();
    current->updateHost(true);
}

void QFlexiWindow::hostMenu ()
{
    QFlexiHost* current = (QFlexiHost*)iv->currentItem();
    current->configure();
}

void QFlexiWindow::hostMenu (QIconViewItem* item, const QPoint& pos)
{
    if (!item) return;
    QFlexiHost* host = (QFlexiHost*)item;
    host->configure();
}

QIconView* QFlexiWindow::iconView () const
{
    return iv;
}

void QFlexiWindow::autoUpdate ()
{
    static unsigned hostCount = 0;
    static QIconViewItem* item = 0;
    if (!updateMenu->isItemChecked(idAutoUpdate))
	return;
    if (hostlistChanged) {
	item = 0;
	hostlistChanged = false;
    }
    if (item)
	item = item->nextItem();
    if (!item) {
	if (0 < hostCount &&
	    hostCount < (1000/UpdateTimerInterval * updateInterval)) {
	    Debug << "Auto-update - waiting" << endl;
	    hostCount++;
	    return;
	}
	item = iv->firstItem();
	hostCount = 0;
    }
    if (!item)
	return;
    Debug << "Auto-updating " << item->text() << endl;
    ((QFlexiHost*)item)->updateHost();
    hostCount++;
}

void QFlexiWindow::update ()
{
    QIconViewItem* item = iv->firstItem();
    while (item != 0) {
	((QFlexiHost*)item)->updateHost(true);
	item = item->nextItem();
    }
    show();
}


void QFlexiWindow::about ()
{
    QMessageBox::about(this,"About QFlexiMon",
		       "QFlexiMon is a Qt-based front end for the\n"
		       "Fleximon flexible monitoring daemon.");
}

void QFlexiWindow::aboutQt ()
{
    QMessageBox::aboutQt(this,"About Qt");
}



//
// Mainprogram
//
int main (int argc, char **argv)
{
    QApplication app (argc,argv);
    app.connect(&app,SIGNAL(lastWindowClosed()),&app,SLOT(quit()));
    QFlexiWindow* fw = new QFlexiWindow ();
    (void)new QFlexiHost(fw->iconView(),"127.0.0.1",FMDefaultPort);
    //(void)new QFlexiHost(fw->iconView(),"192.168.0.1",FMDefaultPort);
    //(void)new QFlexiHost(fw->iconView(),"192.168.0.2",FMDefaultPort);
    fw->setCaption("QFlexiMon");
    fw->show();
    return app.exec();
}
